'''
# coding: utf-8

# # Data & Visualization in Python
'''

'''
To use the `ndarray` and NumPy's other offerings, we need to import it. 
Any script you create that requires NumPy will need to include a line like this:
'''

import numpy as np
x = np.array([3,19,7,11])
y = np.linspace(-2, -1, 25)

from numpy import array, linspace
x = array([0.01, 0.45, -0.3])
y = linspace(0, 1, 50)

'''
Creating arrays
 There are multiple ways to create a numpy array. The first is to convert from a python list. 
'''

arr = np.array([17, -10, 16.8, 11])
arr
type(arr)  # it's an ndarray!
arr.dtype

'''
Numpy arrays have a property of `dtype` which records what is the data type of the arrays members. Notice that the list we passed to `np.array` had all integer values except for one. Numpy arrays are required to have the same data type, so all of the members got converted to floats. 
# 
# The available `dtype`s in NumPy are:
# 
# | Data type	    | Description |
# |---------------|-------------|
# | ``bool_``     | Boolean (True or False) stored as a byte |
# | ``int_``      | Default integer type (same as C ``long``; normally either ``int64`` or ``int32``)| 
# | ``intc``      | Identical to C ``int`` (normally ``int32`` or ``int64``)| 
# | ``intp``      | Integer used for indexing (same as C ``ssize_t``; normally either ``int32`` or ``int64``)| 
# | ``int8``      | Byte (-128 to 127)| 
# | ``int16``     | Integer (-32768 to 32767)|
# | ``int32``     | Integer (-2147483648 to 2147483647)|
# | ``int64``     | Integer (-9223372036854775808 to 9223372036854775807)| 
# | ``uint8``     | Unsigned integer (0 to 255)| 
# | ``uint16``    | Unsigned integer (0 to 65535)| 
# | ``uint32``    | Unsigned integer (0 to 4294967295)| 
# | ``uint64``    | Unsigned integer (0 to 18446744073709551615)| 
# | ``float_``    | Shorthand for ``float64``.| 
# | ``float16``   | Half precision float: sign bit, 5 bits exponent, 10 bits mantissa| 
# | ``float32``   | Single precision float: sign bit, 8 bits exponent, 23 bits mantissa| 
# | ``float64``   | Double precision float: sign bit, 11 bits exponent, 52 bits mantissa| 
# | ``complex_``  | Shorthand for ``complex128``.| 
# | ``complex64`` | Complex number, represented by two 32-bit floats| 
# | ``complex128``| Complex number, represented by two 64-bit floats| 
# 
# #### Arrays from sequences
# 
# We can also create an array that contains a sequence of numbers. To create the range of numbers of 0 to 50, use the `arange` command:
'''

rng = np.arange(0, 51)

# The last number in the range is one less than the value you provided, 
# so we provide 51 to ensure that the last value is 50.
# We can also have the numbers increment by a step size other than 1.

evens = np.arange(0, 23, 2)
quarters = np.arange(0, 1, .25)  # contains 0 to 0.75
tiny = np.arange(0, 1, .01)



# generate 7 numbers between 15 and 19
x = np.linspace(15, 19, 7)

# We can convert an array to a normal python list with the `list()` function:

list(evens)

# The max and min values of an array can be attained with the `amin` and `amax` functions:

np.amin(x)
np.amax(x)

'''
Array arithmetic
# Numeric arrays of the same length can be added, subtracted, multiplied or divide. In NumPy, these operations are defined *element-wise*, in other words, each pair of corresponding elements in the two arrays are operated on, and the result is a new array containing each result.
'''

np.array([1,1,1,1]) + np.array([2,2,2,2])
np.array([1,1,1,1]) - np.array([2,2,2,2])
np.array([1,1,1,1]) * np.array([2,2,2,2])
np.array([1,1,1,1]) / np.array([2,2,2,2])


# Boolean arrays can be "added" or "multiplied" together, which is interpreted as the logical OR / AND respectively:

np.array([True, True, False]) + np.array([True, False, False])
np.array([True, True, False]) * np.array([True, False, False])

# The first example is interpreted as:

# [True OR True, True OR False, False OR False] ==> [True, True, False]
 
# The second example is interpreted as:

# [True AND True, True AND False, False AND False] ==> [True, False, False]

'''
 Mathematical functions on arrays
# In base python, when a function is called on a variable, the function is applied to 
that variable as a whole. For example, to get the absolute value of a number:
'''
    
abs(-2) # ==> 2

# If we had a list of numbers which we wanted the absolute value, we could NOT say:

abs([-2, -7, 1])
# TypeError: bad operand type for abs(): 'list'

# The `TypeError` tells us that `abs()` is not set up to handle lists. However, many functions in NumPy are *vectorized* functions, meaning they can handle a single input or an array of inputs. For example, the absolute value function from the NumPy library is a vectorized function:

np.abs(-3)
np.abs([-2, -7, 1])

# Vectorized functions are not only a convenient shorthand, they are also have better computational performance than the equivalent `for` loop of the same calculation. We will also see vectorized functions in the Pandas library.

'''
Accessing array values
Just like with python lists, we can grab individual elements or a range of elements from an array using square bracket notation:
'''

nums = np.arange(20, 30, .5)
len(nums)
nums[1]   # get the second element
nums[0:3] # get the first three elements

'''
Logical filtering
With a NumPy array, we can use a more sophisticated way of access list elements. Suppose we wanted only the members of `nums` that were larger than 26. First we create a "boolean mask" which will answer the question "is value > 26", returning `True` or `False`, for each value in the `nums` array. The result is a numpy array of booleans:
'''
    
is_gt_26 = nums > 26
is_gt_26
type(is_gt_26)  # another ndarray
is_gt_26.dtype  # data type of array elements is 'bool'
len(is_gt_26)   # same size as nums

# How many values in `nums` are larger than 26? We can make use of how python lets us use booleans as numbers:

sum(is_gt_26)  # 7

# How many values are NOT larger than 26 (which is to say they are less then *or equal to* 26)? We could create anoter mask for this, but it's easier to just take the opposite of each value in `is_gt_26`. In base Python, booleans are negated with the `not` keyword. 

i_like_icecream = True
not i_like_icecream # negate a single boolean value

# Since we are doing something slightly different here -- we want to negate *each member* of the array, and not "negate" the whole array (which doesn't really make sense) -- the `not` keyword will not help us here. NumPy gives us the symbol `~` for taking the opposite of *each value* in a array of booleans, and return a new array of booleans:

~is_gt_26
sum(~is_gt_26)  # 13

# Now we can use the boolean mask to filter our `nums` array:

large_nums = nums[is_gt_26]
len(nums[is_gt_26])

# NumPy lines up the values of `nums` with the `True`/`False` values of `is_gt_26`. If the corresponding mask value is `True`, then it retains the value from `nums`, otherwise it excludes it.

# Typically, we create the mask and filter the array in the same step:

large_nums = nums[nums > 26]

'''
# But it is important to remember that there are a few steps happening here:
# * We create a boolean mask -- either ahead of time or on the fly -- which has the same length as our array
# * The logical statement, for example "`> 26`", is applied to *each* value of `nums`, so the result is an array of `True`/`False` values
# * Our `nums` array and the mask array are then lined up, and the values out of `nums` are filtered based on the corresponding mask value.
'''
'''

# The following checks each value of `nums` for being larger than 21 and less than 24.
'''
np.logical_and(nums > 21, nums < 24)
nums[np.logical_and(nums > 21, nums < 24)]


# and the following checks if at least one of the conditions is true:

np.logical_or(nums < 22, nums > 27)
nums[np.logical_or(nums < 22, nums > 27)]
'''
# Using AND and OR on arrays will become easier with the Pandas library.
# 
# <!-- #### \* More advanced logical condition with np.vectorize()
# we'll just use pandas.apply
'''

def f(x):
     return x**2 > 25

fv = np.vectorize(f)
fv(nums)

# https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.vectorize.html

'''
Two-dimensional arrays
 As the name suggests, N-Dimensional array can have more than one dimension! Multiple dimensions are represented by nesting lists. To create a two dimensional array, (commonly called a *matrix*), we can write the following:
'''
mat = np.array([
        [8, 2, 6, 8],
        [4, 5, 7, 2],
        [3, 9, 7, 1]
       ])
mat
type(mat) # still a ndarray
mat.shape # 3 rows and 4 columns -- returned as a tuple
nrows, ncols = mat.shape
nrows

'''
The `shape` property of an array tells us the size of each of its dimensions.

# To extract a value from the matrix, we use 2-dimesional bracket notation:
'''

mat[1, 3] # 2nd row 4th column

# The first number is the row position and the second is the column position you want.

# To extract an entire row of this matrix, replace the column value with `:`. The colon indicates you want all of the columns. Alternatively, you can specify a range of column positions, which uses normal python list slicing notation:

mat[0,:] # first row 
mat[0, 0:2] # first row and just first 2 columns

# Similarly to extract a single column, replace the row argument with `:` or leave it blank:

mat[:, 2] # 3rd column
mat[1:3, 2] # 3rd column but skipping over the first row
mat[1:3, 2:3] # same as previous, but maintains the vertical structure of the column

'''
Transposing arrays
# Suppose we want to pivot our matrix so that we have 4 rows and 3 columns? This can be done with `transpose`. Each column of `mat` will become a column of `mat2`. 
'''

mat2 = mat.transpose()
mat2

# <!-- not including reshape
# Make sure that the new shape values you give it make sense. What would happen if you tried `mat.reshape((4, 4))`? We get an error, because `mat` has 12 values, but we've asked to convert it to a 4 x 4 matrix, which requires 16 values!
# -->
# 
# <!-- NOPE:
# With a one-dimensional array, it is common to transpose to go from horizontal to vertical format (or vice versa):
x = np.array([.1, .6, .8, .2, .9])
x.shape # 5 by 1
x2 = x.transpose()

'''
Shortcuts for common matrices
# Some matrices are used often enough that there are shortcuts for them. We can make a matrix of all zeros, or all ones as follows:
'''

np.zeros((3,3), dtype=int)
np.ones((5,2), dtype=float)

# We can create a matrix with the same value in every position using `np.full`:

np.full((6,2), -1)

'''
Basic statistics
# NumPy has the ability to perform a handlful of common statistical functions. A sister library of NumPy called [SciPy](https://www.scipy.org/) contains a much broader set of functions for doing statistics. 
# 
# We won't use SciPy, but rather a library called [scikit-learn](http://scikit-learn.org/) which has implementations of advanced machine learning algorithms.
# 
# For now, let's take a look at the statistical functions available in NumPy.
# 
Generating random variables
# 
# ##### Uniform distribution
# NumPy can generate random numbers from different distributions. Whether or not you are familiar with different probability distribution functions, you are certainly familiar with the *uniform* distribution. You use a uniform distribution any time someone asks you to guess a number in a range: any number in that range is equally likely.
# 
# To generate a random number from the uniform distribution, use the following NumPy function:
'''

np.random.uniform(low=0,high=100) # gives us one random number
np.random.uniform(low=0,high=100, size=10) # gives us ten random numbers

# It's important to note that the random numbers generated will never quite reach 100. You may get 99.9999 but never 100. Keep this in mind whenever you work with random numbers.

# Let's test that we're really getting uniformly distributed numbers. Generate many random values from 0 to 100:

y = np.random.uniform(100, 200, 10000) # ten thousand rand. nums in the range [100, 200)

# How many of them are between 100 and 110?

sum(np.logical_and(y > 100 , y <= 110))

# How about between 160 and 170?

sum(np.logical_and(y > 160, y <= 170))

# You should find that in any interval of size 10, the number of `y` values in that range is about the same, no matter what interval you choose. So the `uniform` function seems to be working! Any range of a certain size is equally likely. 

# What is the mean of values we generated? 

np.mean(y)

# We can also compute the median, which is defined as the value of `y` for which half of the values lie below it and the other half lie above it:

np.median(y)

# The mean is also called the 50th percentile. We can compute other percentiles too:

np.percentile(y, 50)  # the median
np.percentile(y, 25)  # the 25th percentile
np.percentile(y, 75)  # the 75th percentile

'''
Normal distribution
# Another common distribution is the normal distribution. Unlike the uniform distribution, where every value is equally likely, values near the center are more likely than values farther from the center. This distribution function has the recognizable "bell curve" shape. You can create variations on the normal distribution by specifying the mean (denoted by $\mu$) and standard deviation (denoted by $\sigma$). The mean controls the location of the center of the bell, and the standard deviation controls its width. The *standard normal* distribution has $\mu = 0$ and $\sigma = 1$.
# ![bell-curve](bell.png)
# *Image source: [Purdue Chemistry](https://www.chem.purdue.edu/courses/chm621/text/stat/common/normal/normal.htm)*
# 
# Let's use NumPy to generate some standard normal random variables:
'''


y = np.random.normal(loc=0, scale=1, size=100000) # lets make 100,000 this time

# (NumPy calls the mean `loc` and the standard deviation `scale`).

# Find the mean and standard deviation of the sample we just created:

np.mean(y)
np.std(y)

# These values will not be exactly 0 and 1. That's because we only have a sample of the distribution! If we could have an infintely large sample, we would get exactly 0 and 1.
# 
# Let's calculate a few percentiles:

np.percentile(y, [25,50,75])

# This says 25% of the values lie below -0.67 or so. Since the normal distribution is symmetric, this also means that another 25% of the values lie above +0.67. 

# Now let's check a very specific percentile:

np.percentile(y, 16)

# You should get numbers a value that is pretty close to -1. This tells us that 16% of the values lie below -1, which due to the symmetry of the distribution function implies that another 16% of the values lie above +1. This leaves 100 - (16 + 16) = 68 percent of values between -1 and +1. We have just learned an important property of the standard deviation for any normal distribution! **68% of the values lie between the mean plus or minus one standard deviation.** In this case, our standard deviation is 1, but this would be true for any value! 
# 
# Let's also calculate these percentiles:

np.percentile(y, [2.5, 1.5])

# You should get values close to -2 and -3. Using the previous reasoning again, 100 - (2.5 + 2.5) = **95% of the values lie within 2 standard deviations of the mean,** and 100 - (.15 + .15) = **99.7% of the values lie withing 3 standard deviations of the mean.**

'''
Data set manipulation with Pandas
# 
# Pandas is a powerful library for cleaning and analyzing data sets in Python. <mark>TODO: say more?</mark>
# 
# Just like we imported the NumPy library, we will do the same for Pandas:
'''

import pandas as pd

# Note: it is not required that you also import `numpy` in order to use `pandas`. 
# However you will often see both of them imported since many projects make use of both.

'''
Series
# The first Pandas object we'll learn is a Series. Think of Series as a NumPy array but with [many additional properties and methods](https://pandas.pydata.org/pandas-docs/stable/generated/pandas.Series.html).
# 
# We can create a series from a normal Python list:
'''

num_series = pd.Series([45, 89, 67, 33])


# In fact, the series values are stored in an `ndarray`! To extract just the values as an `ndarray`, use the `.values` property of the series:

num_series.values

# Pandas supports series of dates, making it a great choice for time series analysis. Date series can be created in a few ways:

# every day between Jan 1 and March 31 2017:
pd.date_range(start='20170101', end='20170331', freq='D')

# Go in intervals of month instead:
pd.date_range(start='20170101', end='20170331', freq='M')

# Not specifying end, but instead the start, freq, and how many periods:
pd.date_range(start='20170101', freq='M', periods=4)

# This function can also create hourly series:

# every hour from Jan 1 midnight to Jan 2 Midnight
pd.date_range(start='20170101', end='20170102', freq='H')
 
# every hour from 8 AM - 1 PM on Jan 1
pd.date_range(start='20170101 08:00', end='20170101 13:00', freq='H')

# every other hour
pd.date_range(start='20170101', end='20170102', freq='2H')

# every 30 minutes
pd.date_range(start='20170101 08:00', end='20170101 13:00', freq='30min')


# To deal just with times, excluding the date value, use `pd.timedelta_range`:

pd.timedelta_range('12:00:00', freq='30S', periods=20)

'''
Series methods
# Series are more powerful than base python lists due to the additional attributes and methods they possess. 
# 
# Suppose we have the following series:
'''

norm_series = pd.Series(np.floor( np.random.normal(50,12,100)))

"""
We've use NumPy to generate 100 normally distributed random variables, 
centered at 50 and with standard deviation 12, and then we've taken 
their "floor" meaning we've truncated any decimal places.
"""

# Some basic attributes and methods of this series are:

norm_series.shape  # number of rows and columns
norm_series.mean()
norm_series.median()
norm_series.std()
norm_series.unique()
norm_series.nunique()
norm_series.argmin()    # position of the minimum value
norm_series.argmax()    # position of the maximum value
norm_series.rank()      # ranks from smallest to largest (rank = 1 for smallest value)
norm_series.rank(ascending=False) # ranks from largest to smallest
norm_series.sort_values()
norm_series.cumsum()  # returns a series that is the cumulative sum of norm_series

# The full list of series attributes and methods is available [here]
# (https://pandas.pydata.org/pandas-docs/stable/generated/pandas.Series.html)
'''
`qcut` function for binning data
# In the section on NumPy we learned about quantiles (synonymous with percentiles). 
# 
# The `qcut` function performs two steps that are often done together on a Series. First it calculates the quantiles that we specify, for example, the 25th, 50th and 75th quantiles. It then creates ranges from these values: (0, q25], (q25, q50], (q50, q75], (q75, 1]. Then it takes every value in your Series, and categories it my which of these ranges it falls into:
'''

norm_q_series = pd.qcut(norm_series, [0, .25, .75, 1])  # note we must supply 0 and 1 in the list!!

# This gives us a new Series, where each value gives the range that the corresponding value in the original Series falls into.
# 
# Let's check to make sure this works:

# double check percentile values using numpy
np.percentile(norm_series, [0,25,50,100])  

# then look at some values and the corresponding quantile label:
norm_series[0] 
norm_q_series[0]

norm_series[12] 
norm_q_series[12]


# > One thing to be aware of: `qcut` uses decimals to specify the quantiles (.5 = 50th percentile), while numpy's `percentile` function uses the full number (50 = 50th percentile). In numpy, .5 would be interpreted as the 0.5th percentile!

'''
DataFrames
# 
# The DataFrame is the single most import object in Pandas. It is a collection of Series of equal lenghts, just like you would have in a spreadsheet. But unlike a spreadsheet, we've seen that Series comes with powerful methods to do many calculations quickly, and a DataFrame itself also has many useful methods.
# 
# <!-- example from here: https://www.nasa.gov/pdf/585035main_ALG_ST_SSA-Altitude.pdf
# https://en.wikipedia.org/wiki/STS-121
# -->
# 
# We create a DataFrame object with the `pd.DataFrame` function, and we specify the Series we want to include, using a dictionary. The dictionary keys become the column names of the data frame, and the dictionary values become the column values.
# 
# For this example, we'll use data from [NASA](https://www.nasa.gov/pdf/585035main_ALG_ST_SSA-Altitude.pdf) containing the initial trajectory of its [STS-121 mission](https://www.nasa.gov/mission_pages/shuttle/shuttlemissions/archives/sts-121.html). It launched on July 4, 2006, and we have a measurement of its altitude recorded every ten seconds after launch.
#'''

times = pd.timedelta_range('14:37:55', freq='10S', periods=13)
heights = pd.Series([7, 938, 4160, 9872, 17635, 26969,
  37746, 50548, 66033, 83966,
  103911, 125512, 147411])

sts121 = pd.DataFrame({'Timestamp': times, 'Altitude': heights})
sts121
type(sts121)
sts121.shape


# We can get some general information about our data frame. The `columns` property returns the column names. The `info()` method gives us some extra info about each column: its data type, and how many null values it has. The `describe()` method computes summary statistics on any numeric field. So, for example, we see the average altitude is 51,900 feet.

sts121.columns
sts121.info()
sts121.describe()

# To extract a column from a DataFrame, there are two possible syntaxes:

sts121['Altitude']
sts121.Altitude  # "object oriented" style

# Both are completely valid, but we have a strong preference for the first way. This avoids confusion with any DataFrame attributes, such as `sts121.columns`, `sts121.shape`, etc.

# The first format also extends to select multiple columns by providing a list of the column names to be extracted:

select_cols = ['Timestamp', 'Altitude']
sts121[select_cols]

# or more compact:
sts121[['Timestamp', 'Altitude']]

# To extract a particular row from a DataFrame, we can use a syntax similar to what we used for `ndarray`s, but with one small change: we must use the `iloc` method. 

final_height = sts121.iloc[12,:]
final_height = sts121.iloc[12] # equivalent without the colon

# Suppose we wanted the height at a specific time though, let's say at `14:38:55`. We would need to figure out what row number this timestamp corresponds to. Luckily we are about to see a much easier way.

# DataFrames in Pandas have a property called the index, which will be of use to us now. When we look at our `sts121` data frame, we see what looks like an unlabeled columns with the numbers 0 to 12. This is the index of the data frame. 

# By default the index is simply the row number (starting with 0). But it can sometimes make sense to use something more descriptive for the index. Let's instead use the timestamp field as the index:

sts121 = sts121.set_index('Timestamp')
sts121
sts121.index


# > There is a slightly more compact alternative using the `inplace` argument. This sets the index of `sts121` without needing to say "`sts121 = `". The following would also set the index:

sts121.set_index('Timestamp', inplace=True)

# Now the rows of our data frame are indexed by the imestamp! (And the Timestamp column has been removed.) This makes it really easy to look up values corresponding to a particular timestamp. To do this we now use the `.loc` method:

sts121.loc['14:38:55']

# Notice we used `loc` not `iloc` like in the first example. The "`i`" in `iloc` stands for integer. We can always use this. As it turns out, the row we wanted was in position 6. So we could also say:

sts121.iloc[6]

# To change the index back to the default, use `reset_index()`, which will change the Index back to `0..12`, and move the `timestamp` values back into the data frame as a column: 

sts121 = sts121.reset_index()
sts121.index

# Next time we make such a data frame, we can specify the index we want from the outset. We need to specify the index on the `heights` Series: 

heights = pd.Series([7, 938, 4160, 9872, 17635, 26969,
 37746, 50548, 66033, 83966,
  103911, 125512, 147411], index=times)  # create an index on the heights Series
  
heights
heights.index

# Now each value of `heights` has a timestamp associated with it. 
# <!-- Maybe we want our data frame to have a timestamp every 1 second instead of ten seconds. We can create a Series for this:

times5sec = pd.timedelta_range(start='14:37:55', end=max(times), freq='5S') 
test = pd.Series(list(np.ones((len(times_1sec)))), index=times_1sec)

sts121 = pd.DataFrame({'Altitude': heights})
sts121


# If we wanted to add additional series to our DataFrame, we would have to make sure they had a compatible index. We can test this by making a dummy series which has an observation every 5 seconds:
# 

times5sec = pd.timedelta_range(start=min(times), end=max(times), freq='5S') 
test = pd.Series(np.ones(len(times5sec)), index=times5sec)

sts121 = pd.DataFrame({'Altitude': heights, 'Test': test})
sts121

'''
Using data from a file
# We've seen how to make a data frame from scratch. Oftentimes though you will want to work with data from a file. A common data format is a csv (comma separated value) file, which is a plain text format (meaning it can be viewed by any simple text editor) having one line per row of data, and each column of data separated by a comma.
# 
# Pandas as a `read_csv` function to import data in this format. In your course materials you should have a csv file called `titanic_train.csv`. We will be using this data set to experiment with various DataFrame functions.
# 
# In addition to csv data, Pandas can read a variety of formats, including Excel, JSON, HTML, Stata, SAS, and even from a SQL connection. The full list of readable and writeable file formats is available [here](http://pandas.pydata.org/pandas-docs/version/0.20/io.html).
# 
# ---------
# 
# Let's read in our csv data:
'''

titanic = pd.read_csv('titanic_train.csv')

# If you receive an error that the file does not exist, you may have to set your working directory. There should be a path listed in the upper right of your screen (above the Variable explorer). Make sure the path shown is the same path where the csv file is saved, and if not, update it.
# 
# What did we just create?

type(titanic)  # a pandas DataFrame!
len(titanic)
nrows, ncols = titanic.shape


# This is a well-known data set containing (the available) information about passengers of the Titanic, such as their name, age and ticket price, and whether or not they survived. We obtained this data from a [Kaggle challenge](https://www.kaggle.com/c/titanic/data), and you can read the documentation there to learn more about each of the fields.

titanic.head() # pulls the first 5 rows (the default is 5)
titanic.head(3) # pull the first 3 rows

# 
# We can view some random rows in the data frame:

titanic.sample(n = 10)  # 10 random rows
titanic.sample(frac = .05) # a random 5% of the rows (891 total rows * 5% = ~44)

# 
# In Spyder, you can also double click on the data frame in the Variable explorer to bring up a view of the data.
# 
# Let's see what's in this data frame.

titanic.columns
titanic.dtypes
titanic.info()

# The output of the `info()` method shows us that in a few cases, the `Embarked` field is null (missing). The `Age` field has several missing values, and the `Cabin` field is null the majority of the time. *We need to be aware of the frequency of missing values if we ever plan on building a model with this data.*

# Since Cabin is missing so often, let's remove it. Removing a column from a data frame makes use of the Python `del` keyword:

del titanic['Cabin']
titanic.head() # Cabin column is gone

# To get an understanding of the range of values in each field we can use `.describe()`:

titanic.describe()

# What is the index of this data frame?

titanic.index


# Maybe we should set the index to be the `PassengerId`. First let's check if this is a good value to use as index:

sum(titanic['PassengerId'].isnull())  # 0, good - it's never missing!
titanic['PassengerId'].nunique()  # 891, good - every value is unique

# So we can set it as the index:

titanic.set_index('PassengerId', inplace=True)
titanic.index # view the new index

'''
# #### 
# What are the possible values of the Survived column? Remember that a DataFrame is just a collection of Series, so if we extract the `'Survived'` series, we can use any of the Series methods we learned earlier:
'''
titanic['Survived'].unique()

# How about Pclass?

titanic['Pclass'].unique()
titanic['Pclass'].nunique() 

# Do this with a few other fields to gain familiarity with the data. Are there any columns whose meaning is not clear to you? (`Embarked` for example). Refer to the documentation linked above to learn the meanings. 
'''
Sorting
# To sort the data frame we can use the `.sort_values()` method. We can specify one or more column names to sort by:
'''

titanic.sort_values(by=['Sex', 'Age'])

sort Sex Z->A, and Age lowest to highest
titanic.sort_values(by=['Sex', 'Age'], ascending=[False, True]) 

# inplace=True means we overwrite titanic with the sorted version of itself
titanic.sort_values(by=['Sex', 'Age'], ascending=[False, True], inplace=True) 
titanic.head()


'''
Filtering
# 
# We learned how to create boolean masks in the second on NumPy. DataFrames also support filtering this way (with the addition of the `loc` method we saw earlier):
'''
age_is_missing = titanic['Age'].isnull()  
# .isnull() returns True when the value is missing (null)
sum(age_is_missing)   # 177 ages are unknown
sum(~age_is_missing)   # 714 ages are known
titanic.loc[age_is_missing,:]
titanic.loc[~age_is_missing,:]

# or more compact:
titanic.loc[titanic['Age'].isnull()]
titanic.loc[~titanic['Age'].isnull()]

# So we can pull all of the rows corresponding to passengers whose age is missing 
# or not missing respectively.

# Pandas provides an even cleaner syntax to filter by column values with the `DataFrame.query()` method. The following finds passengers under the age of 18:

titanic.query('Age < 18')

# This syntax makes it easy to apply more complicated logical filters:

titanic.query('Age < 18 and Sex=="female"')

# *Note that it is still best practice to use the boolean mask method when filtering for missing values (using `.isnull()`).*

# We can use filters to count survivors by gender. We'll query the data frame, which returns a new (filtered) data frame, and then get the row count of that new data frame. The shape attribute of a data frame returns two values -- the number of rows and columns -- so since we just care about the number of rows, we can just take the first value:

titanic.query('Age < 18 and Sex=="female" and Survived==1').shape[0]  # 38
titanic.query('Age < 18 and Sex=="female" and Survived==0').shape[0]  # 17

# This tells us there were 38 female minors who survived, and 17 who didn't. Therefore the survival rate for this cohort is 38 / (38 + 17) = 69%.

# On the other hand if we look and survival rate for males between 25 and 35, the survival rate is much worse:

titanic.query('Age >= 25 and Age < 35 and Sex=="male" and Survived==1').shape[0]  # 31
titanic.query('Age >= 25 and Age < 35 and Sex=="male" and Survived==0').shape[0]  # 107

# Survival rate = 31 / (31 + 107) = 22.5%

# We will create these rates in a more automated way when we look at the `groupby` method, but it is nonetheless a very helpful first step to perform a few manual calculations like this. 
'''
Creating new columns 
# We can easily add new columns to the data frame. Since each data frame column is a series, we need something like
'''
# titanic['new_column_name'] = {some Series}

# If you happen to have a Series already, and the index of that Series is compatible with your data frame, you can add it as follows:

myseries = pd.Series(np.arange(0, 892))
titanic['myseries'] = myseries

# Most of the time though, we want to create a new column based on calculations on existing columns. We could make a new Boolean column called `is_minor`:

is_minor = titanic['Age'] < 18
type(is_minor) # it's a series, so we can add it to the data frame
titanic['is_minor'] = is_minor

# or more compact
titanic['is_minor'] = titanic['Age'] < 18

'''
More complicated column derivations with `Series.apply`
Suppose we wanted to create a new column that indicated the person's title (Mr., Mrs., etc). The Name column includes the title, but it is not presented in a consistent way. Let's make a new column that contains just the title. 

'''

def get_title(full_name):
   if full_name is None:
        # if name is not provided, then we can't provide a title!
       return None
    
    titles = ['Mr.', 'Mrs.','Ms.', 'Miss.', 'Lady.', 'Master',
              'Major.', 'Col.', 'Capt.', 'Rev.', 'Dr.', 'Jonkheer.',
              'Sir.','Mme.', 'Countess', 'Mlle.', 'Don.']
    for title in titles:
        if full_name.find(title) != -1:
            return title
        
# if it makes it to here this means no titles were found in the name
# (or our list of titles is incomplete)
   return '' 


# Let's test our function on a few values before continuing:

get_title('Johnston, Miss. Catherine Helen "Carrie"')
get_title(titanic['Name'][43])


# Now we want to get the title of all 891 passengers. Notice that the following does not work:

get_title(titanic['Name'])

# Our function is built assuming that `full_name` is a string. It can't handle a whole Series being passed to it. Here is when `apply` become useful! `apply` works on a Series and *applies* a function to every value in that Series. So the following is what we need:

titles = titanic['Name'].apply(get_title)
type(titles) # and it gives us back another Series


# So now to put it all together, we can create the title column in our data frame:

titanic['Title'] = titanic['Name'].apply(get_title)
titanic.head()
titanic['Title'].unique()

'''
`groupby` for summarization
# How many passengers on the Titanic were in each class (1st, 2nd, 3rd)? We could answer that with a query:
'''

titanic.query('Pclass==1').shape[0] # 216
titanic.query('Pclass==2').shape[0] # 184
titanic.query('Pclass==3').shape[0] # 491

# But suppose there were many possible values of `Pclass`, this would get quite tedius. When we want to summarize our data by distinct values of a column, we can use `groupby` + an aggregation function. For our example, we would like to group by the `Pclass` variable. The aggregation method we want is to get  the size of each group. We will see examples of other aggregation methods shortly.
# 
# The syntax is as follows:
# ```python
# grouped_df = titanic.groupby('Pclass')
# type(grouped_df)  # Pandas "DataFrameGroupBy Object"
# ```
# This creates a "`DataFrameGroupBy`" object, which tells us that we can now perform an aggregation method on the object, such as `.size()`. 

grouped_df.size()

# This shows one row of aggregated information for each distinct value of the grouping variable `Pclass`, and we confirm the values are correct because we manually calculated them earlier. 

# Other aggregation methods include `sum()`, `mean()`, `std()`, `min()` and `max()`. Let's get the average ticket price (Fare) that people in each class paid:

grouped_df.mean()

# `.mean()` by itself will comput the mean of *all* the variables per group. If we just care about one of them, such as mean Fare, we can write:

grouped_df.mean()['Fare']

# We can get multiple aggregate values at once with the `.agg` method:

grouped_df.agg([np.mean, np.std, np.amax, np.amin])['Fare']
# ```
# Note that we pass the aggregation function itself. 
# This suggests that we could create our own aggregration function:

def filt_min(arr):
# """ returns the min value excluding 0 (which probably denotes 'fare unknown')"""
   return min([a for a in arr if a > 0])
    
grouped_df.agg([np.mean, np.std, np.amax, np.amin, filt_min])['Fare']

# Putting it all together, let's see the number of passengers with various titles, and their average fares:

titanic.groupby('Title').agg([np.size, np.mean])['Fare'].sort_values(by='mean', ascending=False)

# > Remember, whenever you need to, you can always break a complicated process into multiple steps! It's more important to know what each step is doing than to have super compact code. 

'''
Grouping by multiple variables
# What if we wanted to see the breakdwn by the departure port *and* class? There are three ports (C, Q, and S), and three classes. So within each port, we want to see a breakdown by class, resulting in 9 total values. Simply provide a list of column names to the `.groupby` method:
'''
grouped_df = titanic.groupby(['Embarked', 'Pclass'])
summary = grouped_df.agg([np.size, np.mean])['Fare']
summary

# We've calculate the size and mean for each group. 
# For example, the number of 1st class passengers who boarded at port C (Cherbourg) was 85. 
# We can see that the Queenstown port was predominately for third class passengers.
# We can also see that the first class passengers who boarded at 
# Cherbourg paid a higher fare on average than the first class passenegers who boarded at Southampton.

# The summary variable looks a little different than other DataFrames we've seen. 
# This is because it's the first one we've seen with a multi-level index

type(summary)
summary.index

# The multi-level index is extremely useful with aggregate data. 
# To get just the Cherbourg passengers we write:

summary.loc['C']

# To get just the first class passengers from Cherbourg, we write:

summary.loc['C', 1]

# If you prefer having the port names and class numbers as columns in the data frame, as opposed to as indexes, simply employ the `reset_index()` method:

    
summary_noindex = summary.reset_index()
 
# This can be usefule for queries like the following, where you want to incorporate the index variables into a filter:

summary_noindex.query('Embarked in ["Q","S"] and Pclass == 3')

'''
Window functions
# In addition to aggregation over a group, we can also perform what are called 'window functions.' An example will demonstrate what these are. To make this example easier to see, we will consider a subset of the passengers, just men who boarded at Cherbourg, and we'll just select the important columns:
'''

men_ch = titanic.query('Embarked=="C" and Sex=="male"')[['Embarked', 'Pclass', 'Fare']].copy()
men_ch.shape[0]  # 95 values, to make this example more manageable 

# > Note, we want pandas to forget that men_ch came from the titanic data set, and treat it as a completely indepent data set. This is why we used the .copy().

# Suppose we wanted to create a rank value based on the ticket price. We have learned the `.rank()` method for series:

men_ch['Overall_fare_rank'] = men_ch['Fare'].rank(method='first', ascending=False)
# method='first' tells pandas how we want to handle ties

# Double click on the men_ch variable in the Variable explorer to take a look at it. You can click the Fare_rank column header to sort by rank. We've ranked all of the fares in order, regargless of the class.

# But what if we wanted separate ranking systems for first, second, and third class? We would start by creating a grouped data frame:

grouped_df = men_ch.groupby('Pclass')

# Then we apply the rank function:

class_fare_rank = grouped_df.rank(method='first', ascending=False)['Fare']
men_ch['Class_fare_rank'] = class_fare_rank

# Now open the `men_ch` data frame to view the windowed rank. (Note you may want to first sort the data by `Pclass` and `Fare`).

men_ch.sort_values(by=['Pclass','Fare'], ascending=[True, False], inplace=True)

# Other useful window functions include `cumsum`, to compute a cumulative sum, and `shift`, which lets you create lagged or lead series.

'''
 "Wide" vs "Long" data
# 
# A very common task in data wrangling (preparing your analysis), is converting between what is called either "wide" or "long" format. Wide data is much more visually digestable, which is why you're likely to come across it if you are usign data from some type of report. However, long data is much easier for us to work with in Pandas, and generally speaking in most data analysis tools.
# 
# Let's create an example of a wide data set from our Titanic data. Then we'll show how to convert it to long format. Wide data often appears when the values are some type of aggregate. Let's make a data frame with two rows and four columns. The format will be like this:
# 
# |    Sex    |  C  |  Q  |  S  |
# |-----------|-----|-----|-----|
# | `female`  | ... | ... | ... |
# |  `male`   | ... | ... | ... |
# 
# Where the "Q" column in the 2nd row will have the number of men who boarded at the Queenstown port.
# 
# The easiest way will be to make three series:
'''

Q = titanic.query('Embarked=="Q"').groupby('Sex').size()
C = titanic.query('Embarked=="C"').groupby('Sex').size()
S = titanic.query('Embarked=="S"').groupby('Sex').size()
wide = pd.DataFrame({'C':C, 'Q':Q, 'S':S})
wide.reset_index(inplace=True)
wide

# The reason we call this format "wide" is because each value of the Embarked variable has its own column. It makes the information easier to present, but inconvenient to run the various analyses we've been learning. 

# Now let's convert this wide data to the long format. First, a preview of how it should look. The Embarked variable, which had a separate column for each value, should be put into a single column. Therefore the result will have 6 rows:
# 
# |   Sex    | Embarked | pax_count |
# |----------|----------|------|
# | `female` |    C     | ...  |
# | `female` |    Q     | ...  |
# | `female` |    S     | ...  |
# |  `male`  |    C     | ...  |
# |  `male`  |    Q     | ...  |
# |  `male`  |    S     | ...  |
# 
# 
# To convert from our wide data to the long format, we use the Pandas `melt` function. 
# The first argument it takes is the data frame. 
# Then we must specify the variable that corresponds to each row of the wide data set. 
# For use this is the `Sex` variable. 

long = pd.melt(wide, id_vars=['Sex'], var_name='Embarked', value_name='pax_count')
long
# The `var_name` and `value_name` arguments are optional. 
# These let us customize the name that Pandas gives to the newly created columns.

# Compare the wide and long data frames and make sure your understand how they are related. 

# We can convert the `long` data frame back to wide format with the `DataFrame.pivot()` method.

wide2 = long.pivot(index='Sex', columns='Embarked', values='pax_count')

# The `index` argument refers to what values *will become* the index in the new data frame. The `columns` argument says which values will be converted to columns (remember, the different `Embarked` values each get their own column in the wide format). Lastly we supply the `values` argument, which is the field to use to fill in the values of the wide data.
# 
# > Note the slight difference in syntax between `melt`, which is a Pandas function, and `pivot` which is a method of a dataframe. You would say `pd.melt()` but `long.pivot()`. Something you just have to memorize!

'''
Adding more data to a DataFrame via `concat`

Adding rows
# Let's create a few more passengers:
'''

new_passengers = pd.DataFrame({
    'Name': ['Jack Dawson', 'Rose DeWitt Bukater'],
    'Sex': ['male', 'female'],
    'Age': [19, 17],
    'Survived': [0, 1]
    
}, index=[890, 891]) # make sure to give them an index


# Now we can add them to our `titanic` data frame:

titanic = pd.concat([titanic, new_passengers])
titanic.tail

# Notice that we didn't need to provide all of the values for our new pasengers (Fare, Sipsb and several others are not specified) This is not a problem: Pandas will correctly match up the information we do provide.
'''
Adding columns
# 
# If we have two data frames (with compatible indexes), we can create a new data frame by putting the two side by side. To demonstrate this, we'll create two data frames from our titanic data frame, and show how to combine them together:
'''


bio = titanic[['Name','Age','Sex']]
boarding = titanic[['Pclass','Ticket','Fare']]
bio.head()

# To make it interesting, we'll remove some rows from the `boarding` data frame. Let's make boarding just 2nd class passengers.

boarding = boarding.query('Pclass==2')

# To make a new data frame by putting these two side by side, we write:

combined = pd.concat([bio, boarding], axis=1)
combined.head() 

# If we examine the `combined` data frame by opening it in the Variable explorer, we can see that many of the "right hand side" variables are blank (nan). This because for the particular passenger in the "left hand side" data, there was no correspondig value on the right. So Pandas combined them the best it could, by matching the indices, and in the event there was no match, keeping the left side value and putting a blank on the right side. This behavior is called a *left join* and we'll discuss this more in the SQL lecture.

# > Note: We use the `concat` function to both add rows and to add columns. To differentiate between the two we supply the `axis` argumement, which is 0 by default. `axis=0` concatenates by rows and `axix=1` concatenates by columns.
''' 
Merging data frames

# If your two data frames need to be combined based on multiple columns, then you can use `merge`. We'll demonstrate a multi-column merge by creating two data frames.
# 
# Our first data frame will calculate the survival rates for men and women:
'''

gender_size = titanic.groupby('Sex').size()
gender_survivors = titanic.groupby('Sex').sum()['Survived']

gender_total = pd.DataFrame({'total':gender_size, 'survivors': gender_survivors})
gender_total['sur_rate'] = gender_total['survivors'] /gender_total['total']

# The second data frame will calculate survival rates for men and women, and within each of these, by 1st, 2nd, or 3rd class.

gender_class_size = titanic.groupby(['Sex','Pclass']).size()
gender_class_survivors = titanic.groupby(['Sex','Pclass']).sum()['Survived']
gender_class = pd.DataFrame({'total':gender_class_size, 'survivors': gender_class_survivors})

gender_class['sur_rate'] = gender_class['survivors'] / gender_class['total']

# By looking at `gender_total` and `gender_class` we can see that 74% of all women survived, but if you look at the survival rates broken down by class, there is a stark difference between women in 1st-2nd class and 3rd class. Nearly all of the 1st and 2nd class women survived, while only 50% of the 3rd class women survived. You can see how important it is to look at different groupings! 

# Now we can finally see an example of `merge`. Let's combine `gender_total` and `gender_class`. Notice that they have different indexes: `gender_total` has a single index on `Sex`, while `gender_class` has a hierarchical index on `Sex` and `Pclass`. This means `concat` would fall short:

pd.concat([gender_total, gender_class], axis=1) # does not work

# `merge` uses columns not indexes to join, so let's move the index into the data frame:

gender_total.reset_index(inplace=True)
gender_class.reset_index(inplace=True)

# Now we can merge them on the `Sex` column, which they both have in common:

merged = gender_total.merge(gender_class, on="Sex")
merged

# What did this do? The columns that came from gender_total (the "left" table), are suffixed with "`_x`", and the columns that came from `gender_class` are suffixed with "`_y`". We can control the suffixes like this:
merged = gender_total.merge(gender_class, on="Sex", suffixes = ['_gender_all', '_gender_class'])
merged

# The total values are repeated three times, once for each class, which is what we want. Now we can compare each of the within-class values to the gender totals:

compare_rates = merged[['Sex','Pclass','sur_rate_gender_all','sur_rate_gender_class']].copy()
compare_rates['class_difference'] = compare_rates['sur_rate_gender_class'] - compare_rates['sur_rate_gender_all']


# So we can see that 1st and 2nd class women's survival rates were higher than the all-female average by 22.6 and 17.9 percntage points respectively. First class men did better than the all-male average by 17.9 percentage points. Lastly, there was no meaningful difference between 2nd-3rd class men and the all-men average.

'''
Plotting with `matplotlib`
# 
# `matplotlib` is a populat plotting library among scientists and data analysts. It is one of the older python plotting libraries, and for this reason it has become quite flexible and well-documented. Other plotting libraries you make come across are `Seaboarn` (which is built on `matplotlib`), `ggplot` (the Python version of the popular `R` plotting library), `Plotly`, `Bokeh`, and many others. `Pandas` also comes with some plotting capabilities, and these are actually just based on `matplotlib`.
# 
# You can begin to explore the different types of plots you can create with `matplotlib` by browsing their [gallery](https://matplotlib.org/gallery.html). Today we'll learn a few of the basic plot types.
# 
# > Note you may need to change a setting in Spyder in order to more easily work with your plots. Go to Tools > Preferences > IPython Console > Graphics, and make sure in the "Graphics backend" section, the Backend is set to Automatic.
# 
# To use `matploblib` in our script, we import the `pyplot` module of the `matplotlib` package:
'''

import matplotlib.pyplot as plt
'''
Plot types
# #### Scatter plot
# The first plot we'll learn how to make is a scatter plot, which takes an array of `x` values and an array of `y` values. Let's plot the passenger Age vs their Fare for third class passengers:
'''

class3 = titanic.query('Pclass==3')
plt.scatter(class3['Age'], class3['Fare'])

# If we want, we can change the marker from a circle to some other symbol. This is accomplished with the `marker` argument. We'll use "D" for a diamond shaped marker. The full list of marker options is available [here](https://matplotlib.org/api/markers_api.html#module-matplotlib.markers).

plt.scatter(class3['Age'], class3['Fare'], marker="D")

# `matplotlib` has the ability to "map" data values to plot features, such as marker color and size. Let's map the Survived column, which has two values: 0 or 1, to the marker color. To set the color we use `c`:

plt.scatter(class3['Age'], class3['Fare'], marker="D", c=class3['Survived'])

# We could have also provided a single color to be used for all points. The basic options are `b` (blue), `g` (green), `r` (red), `c` (cyan), `m` (magenta), `y` (yellow), `k` (pink), and `w` (white). You can also use a specific color if it's one of the well known named colors, or by providing its RGB code. You can learn more about these options [here](https://matplotlib.org/users/colors.html).

# Now there are some stylistic features we would like to add, such as an `x` and `y` label and a title.

plt.xlabel('Age')
plt.ylabel('Fare')
plt.title('Ticket price vs Age for 3rd class passengers')

'''
Line plot
# 
# The command to make a line plot just uses the `plot` function:
'''

plt.plot([0,1,2, 3], [5, 9, 6, 6])

# Let's make a more interesting plot with our data. We can plot the average survival rate by age group:

# first use qcut to generate age groups
titanic['AgeRange'] = pd.qcut(titanic['Age'], q = 10)

# then group by the AgeRange field and take the average survival rate. 
#  for plotting purposes, we also want the min Age in each group
summ = titanic.query('Pclass==3').groupby('AgeRange').agg([np.mean, np.min])[['Survived', 'Age']]

# make the plot
plt.plot(summ['Age','amin'], summ['Survived']['mean'])

# <!-- [error bar plot?](https://matplotlib.org/examples/statistics/errorbar_demo.html) -->

'''
Histogram
# 
# A histogram is bar blot, in which the height of each bar has been calculated as the number of observations in each range. Suppose we wanted to plot how many ages were in the ranges 0-10, 10-20, 20-30, etc. We can use `plt.hist`, supplied with the values we want to separate each range:
'''

bin_breaks = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]  # or use np.arange()!
plt.hist(titanic[titanic['Age'].notnull()]['Age'],bins=bin_breaks)

# So we can see there are a little over 50 passengers ages 0-10, 100 passengers ages 10-20, etc.

# Alternatively, we can just specify the number of bins we want, and `matplotlib` will compute the breaks for us:
ages = titanic[titanic['Age'].notnull()]['Age']
plt.hist(ages, bins=30) # creates 30 bins


plt.hist(ages, bins=30, histtype='step', cumulative=True) # specify normed=1 to scale the y-axis to [0, 1]

'''
Box plot
# A box plot is a visual summary of the 25th, 50th and 75th percentiles. It also calcultes an upper and lower threshold on what values should be considered "outliers." 
# <!-- https://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.boxplot -->
'''


plt.boxplot(ages)

# The orange line shows the median age, and the top and bottom of the box are the 25th and 75th percentile respectively. The outermost lines are called the "whiskers." These are calculated as follows:

# 1. Compute the *interquartile range* (IQR), which is the difference between the 75th and 25 percentile ($Q75 - Q25$)
# 2. Take the IQR and multiply it by 1.5.
# 3. The lower whisker is computed as $Q25 - 1.5*IQR$, or if this is beyond the smallest data point, then use the value of the smallest data point.
# 4. The upper whisker is computed as $Q75 + 1.5*IQR$, or if this is beyond the largest data point, then use the value of the largest data point.
# 
# Values beyond the whiskeres are considered outliers: they are substaintially far from the rest of the data. 
# 
# You can change to orientation of the plot to horizontal by setting `vert=False`:

plt.boxplot(ages, vert=False)
plt.xlabel('Age')
plt.title('Age Distribution, All Passengers')

# We can put multiple boxplots in the same figure simply by supplying a list of series. The below compares the distribution of ages for those who survived and those who didn't.

surv0 = titanic[titanic['Age'].notnull()].query('Survived==0')
surv1 = titanic[titanic['Age'].notnull()].query('Survived==1')
plt.boxplot([surv0['Age'], surv1['Age']], notch=True)

# Notice we have added the `notch=True` argument. The notch on a boxplot show a cutout around the median. The size of the notch indicates the confidence interval around the median. As a rule of thumb, if the notches of two boxplots do not overlap, then it is very likely there is a statistically significant difference between the two series.

'''
Violin plot

#  - [violin plot?](https://matplotlib.org/examples/statistics/boxplot_vs_violin_demo.html)
#  -
# -->
# 
# ### Multi-plot figures
# We can create figures containing multiple plots, laid out in a grid, using `plt.subplots()`. This command takes the following important arguments:
# <!-- https://matplotlib.org/devdocs/api/_as_gen/matplotlib.pyplot.subplots.html -->
# 
# * `nrows`, `ncols` - how many rows and columns the grid of plots will have. Both are 1 by default.
# * `sharex`, `sharey` - each are a boolean value determining whether the grids should use the same `x` and / or `y` axes. Both are False by default.
# 
# The `subplots` function returns two values, a `Figure` object and a `Axes` object. The `Figure` contains the entire grid and all of the elements inside. The `Axes` is an array, where each member contains a particular subplot.
# 
# For this examples, we'll plot the number of survivors and non-survivors within each class. First let's calculate these numbers with `groupby`:
'''

sur = titanic.groupby(['Pclass', 'Survived']).size()

# Now let's make a figure with three subplots, arranged in a row with three columns. Since they are side-by-side, it makes sense for the y-axis to be shared.

fig, axes = plt.subplots(nrows=1, ncols=3, sharey=True)

# Immediately, a blank figure with three subplots opens. Check what the value of `axes` is. It is an array containing two axes. Now we can control what goes onto each subplot by using the appropriate member of `axes`:

# first class
axes[0].bar([0, 1], sur[1])
axes[0].set_xlabel('First class')

# second class
axes[1].bar([0, 1], sur[2])
axes[1].set_xlabel('Second class')

# third class
axes[2].bar([0, 1], sur[3])
axes[2].set_xlabel('Third class')

plt.suptitle('Survivor count by class')

'''
# Question: what would the plot look like if the `sharey=False` parameter was not specified?
# 
# > Notice there is a different set of functions for editing the x and y labels and the title of a subplot: for a subplot you need to use `set_xlabel`, etc, while for a single plot (from the earlier sections), you can say `plt.xlabel`. This is something you have to remember, although if you only want to learn one syntax , you can treat all of your single plots as subplots with 1 row and 1 column.
# 
# <!-- 
# https://matplotlib.org/examples/pylab_examples/subplots_demo.html
# -->
# 
# 
# ### Customizing the plot
# 
# <!-- https://matplotlib.org/api/axes_api.html -->
# 
# Now we'll see how to modify and style various parts of our plots. For these examples, we'll use the Discovery data. 
'''

seconds = np.arange(0, 130, 10)
heights = pd.Series([7, 938, 4160, 9872, 17635, 26969,
  37746, 50548, 66033, 83966,
  103911, 125512, 147411])
 
sts121 = pd.DataFrame({'Seconds': seconds, 'Altitude': heights})
sts121

# We will use the subplots function to make a 1 x 2 plot. On the left plot we'll show the altitude in feet, and on the right we will transform the altitude to a log scale.


fig, (ax1, ax2) = plt.subplots(nrows=1, ncols=2)
ax1.plot(sts121['Seconds'], sts121['Altitude'])

# In the previous examples, we output the `axes` array from `plt.subplots`. With the above notation we are simply assigning each member of that array to a variable. This saves us a little time: instead of writing `ax[0]` we can refer to `ax1`.
# 
# Now let's make a line with the points marked by specifying a `marker` shape. While we're at it, let's make the line and markers white:

ax1.plot(sts121['Seconds'], sts121['Altitude'], marker='o', color='w')
ax1.set_title('Altitude - Linear Scale')

# We can set the background color of the plot using `ax.set_facecolor`. [Here](https://matplotlib.org/users/colors.html) again is all of the options you have for specifying colors.
# 

ax1.set_facecolor('silver')

# Now a few definitions. The `axes` object contains the `x` and `y` *axis*. (Remember `axes` refers to the whole subplot). Each `x` axis and `y` axis both have *tick marks*, which are the little lines placed a certain intervals along the axis. The `x` axis and `y` axis also have corresponding *tick mark labels*. These are the values just next to or under the ticks. 

# To change the colors of the tick marks we write:

ax1.tick_params('both', color='blue')

# We can also set the color of the axis label as follows:

ax1.xaxis.label.set_color('blue')
ax1.yaxis.label.set_color('blue')

# <!-- Can someone figure out how to change the color of the tick LABELS? the only solution i've found involves calling set_xticklabels, but this means you need to hard code the labels (so when the plot is resized it gets messed up)-->
# 
# Next lets format the `x` tick labels to start with "T +" sign (so T + 20 seconds, etc). We'll make a helper function:

def add_plus(t):
    return 'T +'+str(int(t))

 
# Now we must establish where we want the tick marks to be. `matplotlib` has picked some by default, but the formatting is easier if we choose them ourselves:

# range from the lowest to highest value, in steps of 20 seconds
# note, we add +1 to the end value so that arange includes it
xticks = np.arange(min(sts121['Seconds']), max(sts121['Seconds'])+1, 20)

# Now we set the axis labels to our `xticks` variable:

ax1.set_xticks(xticks)


# And format them with the `'T +'`:

new_xlabels = [add_plus(s) for s in ax1.get_xticks()]

# Finally, we apply the new labels to the graph:

ax1.set_xticklabels(new_xlabels)

# To finish up our `x-axis` on this plot, we should add an *axis label* (not to be confused with the tick label!). There is a nice shortcut we can use if we want to have consistent text formatting (font, size, color) throughout our plot. We can store these in a dictionary:

plot_font = {
        'family': 'Trebuchet MS',
        'color':  'blue',
        'weight': 'normal',
         'size': 16
         }

# And then provide all of these setting in bulk via the `fontdict` parameter:

ax1.set_xlabel('Seconds since liftoff', fontdict=plot_font)

# Now let's format the y-axis. Maybe we should convert the values to be in units of one thousand. Again we make a helper function:

def to_thousands(h):
    return round(h/1000)

# We will put a y tick at every altitude value:

ax1.set_yticks(sts121['Altitude'])

# Now apply the formatting:

new_ylabels = [to_thousands(y) for y in ax1.get_yticks()]
ax1.set_yticklabels(new_ylabels)
ax1.set_ylabel('Altitude (Thousands of Feet)', fontdict=plot_font)

# Now we'll make the second subplot. Recall we used the `sharex=True` argument, which means all the formatting changes we made to the first x-axis will automatically be made to this one. 

ax2.plot(sts121['Seconds'], sts121['Altitude'], marker='o', color='w')
ax2.set_title('Altitude - Log Scale')
ax2.set_facecolor('silver')
ax2.tick_params('both', color='blue')
ax2.xaxis.label.set_color('blue')
ax2.yaxis.label.set_color('blue')
ax2.set_xlabel('Seconds since liftoff', fontdict=plot_font)

# We can use the `semilogy` axes method to transform the y-axis to a log scale:

ax2.set_yscale('log') # convert y axis to log
# ax2.semilogy()  # alternative
ax2.set_ylabel('Altitude (Log Feet)', fontdict=plot_font)

# Finally, we apply a title to the entire plot:

fig.suptitle('Discovery Rocket Launch Altitude after Liftoff',  **plot_font)


plt.plot_date(sts121['Timestamp'], sts121['Altitude'], fmt='-')

# If we want, we can also plot the points on top of the line:

plt.plot_date(sts121['Timestamp'], sts121['Altitude'], fmt='o') # fmt='o' is the default, but we provide it here for clarity



ax2.text(18, 1050, '1,000 feet reached \nin 10 seconds')

# You can insert a line break in an annotation using the `\n` character. Annotations can contain equations made with LaTex notation:

ax1.text(10, 126000, r'$v_e = \sqrt{\frac{2GM}{r}} $')

# The equation should be surrounded by `$` and proceeded by an `r'`. Remeber to use the *original* coordinates, for example, we are only labeling the y value as 126, but the location on the plot is 126,000.

'''
# ### Saving your plots
# The command to save the current plot is 
'''
plt.savefig('discovery.png')

# However, the dimensions that it outputs the plot in are going to be whatever dimensions in which the plot is currently rendered, which might be undesirable. To set specific dimensions, first call the set_size_inches function before saving:

fig.set_size_inches(12, 6)
plt.savefig('discovery.png')

# This way you can be sure you know the dimensions of the saved plot.

